# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import socket
import threading
import pickle

import GameConfig
import Game
import Preferences

# delegate API: 
#     made_move
#     made_ai_move
#     start_ai_move
#     update_ui
#     do_game_over
#     network_connected
#     network_disconnected
#     network_chat
#     network_listening
#     network_listen_failed
#     network_game_started
#     run_on_main_thread

class Controller(object):
    def __init__(self, delegate=None):
        self.game_config = GameConfig.Config()
        self.partial_move = []
        self.highlighted_position = None
        self.delegate = delegate
        
    def call_delegate(self, method, *args, **kwargs):
        if self.delegate and hasattr(self.delegate, method):
            #print 'Calling delegate method:', method
            getattr(self.delegate, method)(self, *args, **kwargs)
            
    def run_on_main_thread(self, action):
        if self.delegate and hasattr(self.delegate, 'run_on_main_thread'):
            self.delegate.run_on_main_thread(action)
        else:
            action()
            
    def call_delegate_on_main_thread(self, method, *args, **kwargs):
        if self.delegate and hasattr(self.delegate, method):
            self.run_on_main_thread(lambda: getattr(self.delegate, method)(self, *args, **kwargs))
        
    def game(self):
        return self.game_config.game

    def current_game_state(self):
        if not self.game(): return None
        return self.game().current_state

    def start_game(self):
        # assumes GameConfig.select_game has already been called
        if self.game_config.network_players():
            self.broadcast_new_game()
        self.made_move() # to force UI update and AI turn
        
    def is_human_turn(self):
        return (not self.is_game_over()) and self.game_config.current_player_type()==GameConfig.HUMAN_PLAYER
        
    def is_ai_turn(self):
        return (not self.is_game_over()) and self.game_config.current_player_type()==GameConfig.AI_PLAYER

    def is_network_turn(self):
        return (not self.is_game_over()) and self.game_config.current_player_type()==GameConfig.NETWORK_PLAYER
        
    def is_game_over(self):
        return self.current_game_state().is_game_over()
        
    def made_move(self, move=None, prevstate=None):
        animations = prevstate.animations_for_move(move, self.current_game_state()) if (move and prevstate) else []
        self.call_delegate('made_move', prevstate, animations)
        if self.current_game_state().is_game_over():
            self.call_delegate('do_game_over')
        else:
            if self.is_ai_turn():
                self.call_delegate('start_ai_move')

    def make_human_move(self, move):
        prevstate = self.current_game_state()
        self.game().make_move(move)
        self.clear_highlighted_positions()
        for i,c in self.game_config.network_players():
            c.send_object({'type': 'move', 'move': move})
        self.made_move(move, prevstate)

    def clear_highlighted_positions(self):
        self.partial_move = []
        self.highlighted_position = None
        
    def do_ai_move(self, threaded=False):
        def go():
            prevstate = self.current_game_state()
            move, newstate, nmoves, t = self.game().make_ai_move(self.game_config.current_ai())
            # don't call made_move here because the delegate may need to update before the next move
            # the delegate should call made_move to continue 
            self.call_delegate_on_main_thread('made_ai_move', prevstate, prevstate.animations_for_move(move, newstate))
            
        if threaded:
            threading.Thread(target=go).start()
        else:
            go()
        
    def highlight_position_if_valid(self, pos):
        """ If pos is a legal partial or full move, highlights it and returns True. If not, clears highlight and returns False """
        next_highlight = None
        if pos and self.current_game_state().valid_partial_move(tuple(self.partial_move + [pos])):
            next_highlight = pos
            
        if self.highlighted_position!=next_highlight:
            self.highlighted_position = next_highlight
            self.call_delegate('update_ui')
            
        return (next_highlight is not None)

    def highlight_next_valid_position(self, direction=+1):
        if not self.is_human_turn(): return self.highlighted_position
        d = 1 if direction>0 else -1
        # get the coord index to start looking forward or backward from
        starti = 0
        indexes = list(self.current_game_state().index_array().flat)
        ni = len(indexes)
        if self.highlighted_position:
            starti = (indexes.index(self.highlighted_position)+d) % ni
        i = starti
        # go forward or back until we find a good partial (or total) move
        found = None
        n = 0
        while not found and n<ni:
            n += 1
            if self.current_game_state().valid_partial_move(tuple(self.partial_move + [indexes[i]])):
                found = indexes[i]
            i = (i+d)%ni            
        self.highlighted_position = found
        self.call_delegate('update_ui')
        return self.highlighted_position

    def select_position(self, pos, automove=True):
        if not self.is_human_turn(): return
        if pos:
            self.partial_move.append(pos)
            #print self.partial_move
            self.highlighted_cell = None
            if automove and self.current_game_state().valid_move(self.partial_move):
                move = tuple(self.partial_move)
                self.make_human_move(move)
            else:
                self.call_delegate('update_ui')

    def undo_partial_move(self):
        if self.partial_move:
            self.partial_move = self.partial_move[:-1]
            self.call_delegate('update_ui')
            
    def can_human_pass(self):
        return self.is_human_turn() and self.current_game_state().valid_move(())
    
    def make_human_pass(self):
        self.make_human_move(())
        
    def can_undo_move(self):
        # if game is over, allow undo if there's a human player
        if self.is_game_over(): 
            return any(GameConfig.HUMAN_PLAYER==self.game_config.player_type(i) for i in xrange(1, 1+self.current_game_state().num_players))
        # don't undo if only 1 prior move and it was AI
        if self.game().move_count()==1 and self.game_config.player_type(1): 
            return False
        return self.is_human_turn() and self.game().can_undo_move()
        
    def undo_move(self, to_human=True):
        self.game().undo_move()
        while to_human and (not self.is_human_turn()) and self.game().can_undo_move():
            self.game().undo_move()
        self.made_move()
        
    def can_redo_move(self):
        return self.is_human_turn() and self.game().can_redo_move()
        
    def redo_move(self, to_human=True):
        self.game().redo_move()
        while to_human and (not self.is_human_turn()) and self.game().can_redo_move():
            self.game().redo_move()
        self.made_move()
          
    def score_label(self, pnum):
        if self.current_game_state().has_scores() and self.current_game_state().num_players>=pnum:
            return str(self.current_game_state().player_score(pnum)) 
        return ""
        
    def player_label_text(self):
        username = Preferences.prefs.get_username()
        ptypes = [self.game_config.player_type(1), self.game_config.player_type(2)]
        if self.current_game_state().num_players==1:
            return (username, "") if ptypes[0]==GameConfig.HUMAN_PLAYER else ("Computer", "")
        
        network_players = self.game_config.network_players()
        if network_players:
            # assume only one
            pnum, chat = network_players[0]
            netplayer_name = chat.player_name if chat.player_name else "Opponent"
            return (username, netplayer_name) if pnum==2 else (netplayer_name, username)
        
        if ptypes[0]==ptypes[1]==GameConfig.HUMAN_PLAYER:
            return ("Player 1", "Player 2")
        if ptypes[0]==ptypes[1]==GameConfig.AI_PLAYER:
            return ("Computer 1", "Computer 2")
            
        if ptypes[0]==GameConfig.HUMAN_PLAYER:
            return (username, "Computer")
        else:
            return ("Computer", username)        
        
    def game_status_message(self):
        # FIXME: need more flexible messages, method on GameState? also strings should be in resource file
        if self.is_game_over():
            if self.current_game_state().num_players==1: return "Game Over"
            winner = self.current_game_state().winning_player()
            if winner==0: return "The game is a draw"
            return self.player_label_text()[winner-1] + " wins"
        return ''
        
    # load/save
    def save_game(self, path):
        game_dict = {'game_config': self.game_config.serialize(), 'game_data': self.game().serialize(), 'version': 1}
        pickle.dump(game_dict, open(path, 'w'))

    def load_game(self, path):
        game_dict = pickle.load(open(path))
        self.game_config.update_from_dict(game_dict['game_config'])
        self.game_config.game = Game.deserialize_game(game_dict['game_data'])
            
    # networking
    def network_connect(self, host, port):
        def connectInThread():
            try:
                sock = socket.socket()
                sock.connect( (host, port) )
                #print 'Got connection'
                self.setup_network_connection(sock, 1, False)
            except Exception, ex:
                self.call_delegate_on_main_thread('network_connect_failed', ex)
                
        t = threading.Thread(target=connectInThread)
        t.setDaemon(True)
        t.start()
        
    def network_listen(self, port):
        def acceptConnectionInThread():
            try:
                serverSocket = socket.socket()
                serverSocket.bind(('', port))
                serverSocket.listen(1)
                self.call_delegate_on_main_thread('network_listening', port, serverSocket)

                cs, addr = serverSocket.accept()
                # check abort flag which can be set in abort_network_listen
                if getattr(self, 'abort_listen', False):
                    self.abort_listen = False
                    serverSocket.close()
                    return
                    
                self.setup_network_connection(cs, 2, True)
            except Exception, ex:
                import traceback
                traceback.print_exc()
                self.call_delegate_on_main_thread('network_listen_failed', port, ex)
        
        t = threading.Thread(target=acceptConnectionInThread)
        t.setDaemon(True)
        t.start()
        
    # bit of a hack, connect to the local listening socket and close immediately, setting a flag to tell the server thread to abort
    def abort_network_listen(self, port):
        try:
            self.abort_listen = True
            sock = socket.socket()
            sock.connect( ('localhost', port) )
            sock.close()
        except:
            pass
            
    def setup_network_connection(self, sock, netplayer, isserver):
        from PGChat import PGChat
        chat = PGChat(sock, self)
        self.game_config.set_network_for_player(chat, netplayer)
        chat.run()
        # clear AIs and notify delegate
        self.game_config.clear_ais()
        self.call_delegate_on_main_thread('network_connected', isserver, sock)
        return chat
        
    def disconnect_network(self):
        for pnum,chat in self.game_config.network_players():
            try:
                chat.push(None)
            except:
                pass
            self.game_config.set_network_for_player(None, pnum)
        
    def handle_network_disconnect(self):
        self.disconnect_network()
        self.call_delegate('network_disconnected')
        
    def handle_network_message(self, chat, msg):
        def doit():
            #print 'Got network message:',msg
            if msg['type']=='new':
                self.game_config.update_from_dict(msg['config'])
                self.game_config.game = Game.deserialize_game(msg['game'])
                self.call_delegate('network_game_started')
                self.made_move()
            elif msg['type']=='move':
                if chat==self.game_config.current_network():
                    prevstate = self.current_game_state()
                    self.game().make_move(msg['move'])
                    self.made_move(msg['move'], prevstate)
                else:
                    print 'Move received from network out of turn!'
            elif msg['type']=='chat':
                self.call_delegate('network_chat', chat, msg.get('chat'))
            elif msg['type']=='name':
                chat.player_name = msg.get('name')
                self.call_delegate('update_ui')
            else:
                print 'Unknown message', msg
        
        self.run_on_main_thread(doit)
    
    def broadcast_network_message(self, msg):
        for pnum,chat in self.game_config.network_players():
            chat.send_object(msg)
        
    def broadcast_new_game(self):
        msg = {'type': 'new', 'game': self.game().serialize(), 'config': self.game_config.serialize()}
        self.broadcast_network_message(msg)
    
    def broadcast_player_name(self, name=None):
        if name is None:
            name = Preferences.prefs.get_username()
        self.broadcast_network_message({'type': 'name', 'name': name})
            
    def broadcast_chat_message(self, chat_message):
        self.broadcast_network_message({'type': 'chat', 'chat': chat_message})
        